import { HapClient, HapMonitor } from "../HapClient";
import { log } from '../Types/types';
import { singleton, inject, container } from 'tsyringe';
import { IServiceType, ICharacteristicType } from "../HapClient/interfaces";
import { IConfig } from "../Models/IConfig";
import { Database } from "sqlite3";
import { mkdirSync } from 'fs';
import { join } from 'path';
import Knex from 'knex';
import { IEvent } from "../Models/db/iEvent";
import { IEventServiceInstance } from "../Models/db/iEventServiceInstance";
import { IEventInstanceCharacteristics } from "../Models/db/iEventServiceInstanceCharacteristics";
import { ISequence } from "../Models/db/iSequence";


@singleton()
export default class Monitor {
    private log: log;
    private config: IConfig
    private client: HapClient;
    private pin = "031-45-154";
    private databaseLocation = `/Users/brandonwatson/.homebridge/automation`;
    private databaseName = 'automation.db';
    private db?: Database;
    private knex?: Knex;
    private hapMonitor?: HapMonitor;
    private services: Array<IServiceType> = [];

    constructor(@inject("log") log: log, @inject("config") config: IConfig) {
        this.log = log;
        this.config = config;

        this.client = new HapClient({
            pin: this.pin,
            logger: this.log,
            config: {}
        });

        //Overwrite default db location
        if (config.databaseLocation) {
            this.databaseLocation = config.databaseLocation;
        }

        this.initializeDatabase().catch((err) => {
            this.log(`Error connecting to sqlite database: ${err}`);
        });

        this.initializeHapMonitor();
    }

    private get dbFullPath(): string {
        return join(this.databaseLocation, this.databaseName);
    }

    private initializeHapMonitor = async (): Promise<void> => {
        await this.client.discover();
        this.services = await this.client.getAllServices();

        this.hapMonitor = new HapMonitor(this.log, this.log, this.pin, this.services);

        //Set service updated event
        this.hapMonitor.on('service-updated', this.onServiceUpdated);

        this.log("Monitor ready");
    }

    private initializeDatabase = async (): Promise<void> => {
        try {
            mkdirSync(this.databaseLocation);
        } catch (err) {
            if (err.code !== 'EEXIST') {
                throw err;
            }
        }

        try {
            await this.initDbFile(this.dbFullPath);
        } catch (err) {
            throw err;
        }

        this.knex = Knex({
            client: 'sqlite3',
            connection: {
                filename: this.dbFullPath
            }
        });

        container.register<Knex>('knex', { useValue: this.knex });

        const hasTables = await this.knex.schema.hasTable('events');

        if (!hasTables) {
            //create databse tables
            await this.knex.raw(`CREATE TABLE "event_instance_characteristics" (
                "id"	INTEGER NOT NULL PRIMARY KEY AUTOINCREMENT UNIQUE,
                "instanceId"	INTEGER NOT NULL,
                "key"	TEXT NOT NULL,
                "type"	TEXT NOT NULL,
                "value"	BLOB NOT NULL)`);

            await this.knex.raw(`CREATE TABLE "event_service_instances" (
                "id"	INTEGER NOT NULL PRIMARY KEY AUTOINCREMENT UNIQUE,
                "eventId"	INTEGER NOT NULL,
                "serviceUniqueId"	INTEGER NOT NULL)`);

            await this.knex.raw(`CREATE TABLE "events" (
                "id"	INTEGER NOT NULL PRIMARY KEY AUTOINCREMENT UNIQUE,
                "timestamp"	BLOB NOT NULL)`);
        }
    };

    private initDbFile = (path: string): Promise<Database> => {
        return new Promise((resolve, reject): void => {
            const db: Database = new Database(path, (err): void => {
                if (err) {
                    return reject(err);
                } else {
                    return resolve(db);
                }
            });
        });
    }

    private onServiceUpdated = async (updatedServices: Array<IServiceType>): Promise<void> => {
        this.knex && await this.knex('events').insert<IEvent>({
            timestamp: this.knex && this.knex.fn.now(),
        });
        const seq = this.knex && await this.knex<ISequence>('sqlite_sequence').where('name', 'events').first();
        const eventId = seq && seq.seq;


        await Promise.all(updatedServices.map(async (service: IServiceType): Promise<void> => {
            this.knex && await this.knex('event_service_instances').insert<IEventServiceInstance>({
                eventId: eventId,
                serviceUniqueId: service.uniqueId
            });

            const instanceSequence = this.knex && await this.knex<ISequence>('sqlite_sequence').where('name', 'event_service_instances').first();
            const instanceId = instanceSequence && instanceSequence.seq;

            await Promise.all(service.serviceCharacteristics.map(async (characteristic: ICharacteristicType): Promise<void> => {
                this.knex && await this.knex('event_instance_characteristics').insert<IEventInstanceCharacteristics>({
                    instanceId: instanceId,
                    key: characteristic.serviceName,
                    type: characteristic.type,
                    value: characteristic.value
                });
            }));
        }));

    }
}